class LayerPreprocess(object):
    """
    Applies the `preprocess` function of a convolutional Layer to the adjacency
    matrix.

    **Arguments**

    - `layer_class`: the class of a layer from `spektral.layers.convolutional`,
    or any Layer that implements a `preprocess(adj)` method.
    """
    def __init__(self, layer_class):
        self.layer_class = layer_class

    def __call__(self, graph):
        if graph.a is not None and hasattr(self.layer_class, 'preprocess'):
            graph.a = self.layer_class.preprocess(graph.a)

        return graph



from spektral.layers.ops import sp_matrix_to_sp_tensor


class AdjToSpTensor(object):
    """
    Converts the adjacency matrix to a SparseTensor.
    """
    def __call__(self, graph):
        if graph.a is not None:
            graph.a = sp_matrix_to_sp_tensor(graph.a)

        return graph



import os.path as osp

import numpy as np
import scipy.sparse as sp
from scipy.io import loadmat
from tensorflow.keras.utils import get_file

from spektral.data import Dataset, Graph


class QM7(Dataset):
    """
    The QM7b dataset of molecules from the paper:

    > [MoleculeNet: A Benchmark for Molecular Machine Learning](https://arxiv.org/abs/1703.00564)<br>
    > Zhenqin Wu et al.

    The dataset has no node features.
    Edges and edge features are obtained from the Coulomb matrices of the
    molecules.

    Each graph has a 14-dimensional label for regression.
    """
    url = 'http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/qm7b.mat'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def download(self):
        get_file('qm7b.mat', self.url, extract=True, cache_dir=self.path,
                 cache_subdir=self.path)

    def read(self):
        print('Loading QM7 dataset.')
        mat_file = osp.join(self.path, 'qm7b.mat')
        data = loadmat(mat_file)

        coulomb_matrices = data['X']
        labels = data['T']

        output = []
        for i in range(len(coulomb_matrices)):
            row, col, data = sp.find(coulomb_matrices[i])
            a = sp.csr_matrix((np.ones_like(data), (row, col)))
            e = data[:, None]
            y = labels[i]
            output.append(Graph(a=a, e=e, y=y))

        return output



import os
import os.path as osp

import numpy as np
import scipy.sparse as sp
from tensorflow.keras.utils import get_file

from spektral.data import Dataset, Graph
from spektral.utils import label_to_one_hot
from spektral.utils.io import load_csv, load_sdf

ATOM_TYPES = [1, 6, 7, 8, 9]
BOND_TYPES = [1, 2, 3, 4]


class QM9(Dataset):
    """
    The QM9 chemical data set of small molecules.

    In this dataset, nodes represent atoms and edges represent chemical bonds.
    There are 5 possible atom types (H, C, N, O, F) and 4 bond types (single,
    double, triple, aromatic).

    Node features represent the chemical properties of each atom and include:

    - The atomic number, one-hot encoded;
    - The atom's position in the X, Y, and Z dimensions;
    - The atomic charge;
    - The mass difference from the monoisotope;

    The edge features represent the type of chemical bond between two atoms,
    one-hot encoded.

    Each graph has an 18-dimensional label for regression.

    **Arguments**

    - `amount`: int, load this many molecules instead of the full dataset
    (useful for debugging).
    """
    url = 'https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb9.tar.gz'

    def __init__(self, amount=None, **kwargs):
        self.amount = amount
        super().__init__(**kwargs)

    def download(self):
        get_file('qm9.tar.gz', self.url, extract=True, cache_dir=self.path,
                 cache_subdir=self.path)
        os.remove(osp.join(self.path, 'qm9.tar.gz'))

    def read(self):
        print('Loading QM9 dataset.')
        sdf_file = osp.join(self.path, 'gdb9.sdf')
        data = load_sdf(sdf_file, amount=self.amount)  # Internal SDF format

        x_list, a_list, e_list = [], [], []
        for mol in data:
            x = np.array([atom_to_feature(atom) for atom in mol['atoms']])
            a, e = mol_to_adj(mol)
            x_list += [x]
            a_list += [a]
            e_list += [e]

        # Load labels
        labels_file = osp.join(self.path, 'gdb9.sdf.csv')
        labels = load_csv(labels_file)
        labels = labels.set_index('mol_id').values[:, 1:]
        if self.amount is not None:
            labels = labels[:self.amount]

        return [Graph(x=x, a=a, e=e, y=y)
                for x, a, e, y in zip(x_list, a_list, e_list, labels)]


def atom_to_feature(atom):
    atomic_num = label_to_one_hot(atom['atomic_num'], ATOM_TYPES)
    coords = atom['coords']
    charge = atom['charge']
    iso = atom['iso']

    return np.concatenate((atomic_num, coords, [charge, iso]), -1)


def mol_to_adj(mol):
    row, col, edge_attr = [], [], []
    for bond in mol['bonds']:
        start, end = bond['start_atom'], bond['end_atom']
        row += [start, end]
        col += [end, start]
        edge_attr += [bond['type']] * 2

    a = sp.csr_matrix((np.ones_like(row), (row, col)))
    edge_attr = np.array([label_to_one_hot(e, BOND_TYPES)
                          for e in edge_attr])
    return a, edge_attr



import numpy as np
import scipy.sparse as sp

from spektral.data import Dataset, Graph


class OGB(Dataset):
    """
    Wrapper for datasets from the [Open Graph Benchmark (OGB)](https://ogb.stanford.edu/).

    **Arguments**

    - `dataset`: an OGB library-agnostic dataset.

    """
    def __init__(self, dataset, **kwargs):
        self.dataset = dataset
        super().__init__(**kwargs)

    def read(self):
        if len(self.dataset) > 1:
            return [Graph(*_elem_to_numpy(elem)) for elem in self.dataset]
        else:
            # OGB crashed if we try to iterate over a NodePropPredDataset
            return [Graph(*_elem_to_numpy(self.dataset[0]))]


def _elem_to_numpy(elem):
    graph, label = elem
    n = graph['num_nodes']
    x = graph['node_feat']
    row, col = graph['edge_index']
    a = sp.csr_matrix((np.ones_like(row), (row, col)), shape=(n, n)).tocsr()
    e = graph['edge_feat']

    return x, a, e, label



from tensorflow.keras import backend as K
from tensorflow.keras.layers import Dense

from spektral.layers.convolutional.message_passing import MessagePassing
from spektral.utils import normalized_adjacency


class TAGConv(MessagePassing):
    r"""
    A Topology Adaptive Graph Convolutional layer (TAG) from the paper

    > [Topology Adaptive Graph Convolutional Networks](https://arxiv.org/abs/1710.10370)<br>
    > Jian Du et al.

    **Mode**: single, disjoint.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \Z = \sum\limits_{k=0}^{K} \D^{-1/2}\A^k\D^{-1/2}\X\W^{(k)}
    $$

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `K`: the order of the layer (i.e., the layer will consider a K-hop
    neighbourhood for each node);
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(self,
                 channels,
                 K=3,
                 aggregate='sum',
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super().__init__(aggregate=aggregate,
                         activation=activation,
                         use_bias=use_bias,
                         kernel_initializer=kernel_initializer,
                         bias_initializer=bias_initializer,
                         kernel_regularizer=kernel_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         kernel_constraint=kernel_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)
        self.channels = channels
        self.K = K
        self.linear = Dense(channels,
                            activation=activation,
                            use_bias=use_bias,
                            kernel_initializer=kernel_initializer,
                            bias_initializer=bias_initializer,
                            kernel_regularizer=kernel_regularizer,
                            bias_regularizer=bias_regularizer,
                            activity_regularizer=activity_regularizer,
                            kernel_constraint=kernel_constraint,
                            bias_constraint=bias_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        edge_weight = a.values

        output = [x]
        for k in range(self.K):
            output.append(self.propagate(x, a, edge_weight=edge_weight))
        output = K.concatenate(output)

        return self.linear(output)

    def message(self, x, edge_weight=None):
        x_j = self.get_j(x)
        return edge_weight[:, None] * x_j

    @property
    def config(self):
        return{
            'channels': self.channels,
        }

    @staticmethod
    def preprocess(a):
        return normalized_adjacency(a)



from .diff_pool import DiffPool
from .global_pool import (GlobalSumPool, GlobalAvgPool, GlobalMaxPool,
                          GlobalAttnSumPool, GlobalAttentionPool, SortPool)
from .mincut_pool import MinCutPool
from .sag_pool import SAGPool
from .topk_pool import TopKPool



import copy
import os.path as osp

import numpy as np
import tensorflow as tf

from spektral.data import Graph
from spektral.data.utils import get_spec
from spektral.datasets.utils import DATASET_FOLDER


class Dataset:
    """
    A container for Graph objects. This class can be extended to represent a
    graph dataset.

    Datasets can be accessed with indices (`dataset[0]` returns a `Graph`),
    iterables  (`dataset[[1, 2, 3]]` returns a `Dataset`) or slices
    (`dataset[start:stop]` also returns a `Dataset`).
    They can also be shuffled (`np.random.shuffle(dataset)` shuffles in-place),
    and iterated over (`for graph in dataset: ...`).

    They should generally behave like Numpy arrays for any operation that uses
    simple 1D indexing.

    Datasets have the following properties that automatically computed from the
    graphs:

        - `n_nodes`: the number of nodes in the dataset (returns `None` if the
        number changes between graphs);
        - `n_node_features`: the size of the node features (returns `None` if
        the size changes between graphs or is not defined);
        - `n_edge_features`: the size of the edge features (returns `None` if
        the size changes between graphs or is not defined);
        - `n_labels`: the size of the labels (returns `None` if the size changes
        between graphs or is not defined); this is computed as the innermost
        dimension of the labels (i.e., `y.shape[-1]`).

    Any additional `kwargs` passed to the constructor will be automatically
    assigned as instance attributes of the dataset.

    Datasets also offer three main manipulation functions to apply callables to
    their graphs:

    - `apply(transform)`: replaces each graph with the output of
    `transform(graph)`. This should always be a `Graph` object, although no
    checks are made to ensure it (to give you more flexibility). See
    `spektral.transforms` for some ready-to-use transforms.
    For example: `apply(spektral.transforms.NormalizeAdj())` normalizes the
    adjacency matrix of each graph in the dataset.
    - `map(transform, reduce=None)`: returns a list containing the output
    of `transform(graph)` for each graph. If `reduce` is a `callable`, then
    returns `reduce(output_list)` instead of just `output_list`.
    For instance: `map(lambda: g.n_nodes, reduce=np.mean)` will return the
    average number of nodes in the dataset.
    - `filter(function)`: removes from the dataset any graph for which
    `function(graph)` returns `False`.
    For example: `filter(lambda: g.n_nodes < 100)` removes from the dataset all
    graphs bigger than 100 nodes.

    You can extend this class to create your own dataset.
    To create a `Dataset`, you must implement the `Dataset.read()` method, which
    must return a list of `spektral.data.Graph` objects, e.g.,

    ```
    class MyDataset(Dataset):
        def read(self):
            return [Graph(x=x, adj=adj, y=y) for x, adj, y in some_magic_list]
    ```

    The class also offers a `download()` method that is automatically called
    if the path returned by the `Dataset.path` attribute does not exists.
    This defaults to `~/.spektral/datasets/ClassName/`.

    You can implement this however you like, knowing that `download()` will be
    called before `read()`. You can also override the `path` attribute to
    whatever fits your needs.

    Have a look at the `spektral.datasets` module for examples of popular
    datasets already implemented.

    **Arguments**

    - `transforms`: a callable or list of callables that are automatically
    applied to the graphs after loading the dataset.
    """
    def __init__(self, transforms=None, **kwargs):

        # Read extra kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

        # Download data
        if not osp.exists(self.path):
            self.download()

        # Read graphs
        self.graphs = self.read()
        if len(self.graphs) == 0:
            raise ValueError('Datasets cannot be empty')

        # Apply transforms
        if transforms is not None:
            if not isinstance(transforms, (list, tuple)) and callable(transforms):
                transforms = [transforms]
            elif not all([callable(t) for t in transforms]):
                raise ValueError('`transforms` must be a callable or list of '
                                 'callables')
            else:
                pass
            for t in transforms:
                self.apply(t)

    def read(self):
        raise NotImplementedError

    def download(self):
        pass

    def apply(self, transform):
        if not callable(transform):
            raise ValueError('`transform` must be callable')

        for i in range(len(self.graphs)):
            self.graphs[i] = transform(self.graphs[i])

    def map(self, transform, reduce=None):
        if not callable(transform):
            raise ValueError('`transform` must be callable')
        if reduce is not None and not callable(reduce):
            raise ValueError('`reduce` must be callable')

        out = [transform(g) for g in self.graphs]
        return reduce(out) if reduce is not None else out

    def filter(self, function):
        if not callable(function):
            raise ValueError('`function` must be callable')
        self.graphs = [g for g in self.graphs if function(g)]

    def __getitem__(self, key):
        if not (np.issubdtype(type(key), np.integer) or
                isinstance(key, (slice, list, tuple, np.ndarray))):
            raise ValueError('Unsupported key type: {}'.format(type(key)))
        if np.issubdtype(type(key), np.integer):
            return self.graphs[int(key)]
        else:
            dataset = copy.copy(self)
            if isinstance(key, slice):
                dataset.graphs = self.graphs[key]
            else:
                dataset.graphs = [self.graphs[i] for i in key]
            return dataset

    def __setitem__(self, key, value):
        is_iterable = isinstance(value, (list, tuple))
        if not isinstance(value, (Graph, list, tuple)):
            raise ValueError('Datasets can only be assigned Graphs or '
                             'sequences of Graphs')
        if is_iterable and not all([isinstance(v, Graph) for v in value]):
            raise ValueError('Assigned sequence must contain only Graphs')
        if is_iterable and isinstance(key, int):
            raise ValueError('Cannot assign multiple Graphs to one location')
        if not is_iterable and isinstance(key, (slice, list, tuple)):
            raise ValueError('Cannot assign one Graph to multiple locations')
        if not (isinstance(key, (int, slice, list, tuple))):
            raise ValueError('Unsupported key type: {}'.format(type(key)))

        if isinstance(key, int):
            self.graphs[key] = value
        else:
            if isinstance(key, slice):
                self.graphs[key] = value
            else:
                for i, k in enumerate(key):
                    self.graphs[k] = value[i]

    def __len__(self):
        return len(self.graphs)

    def __repr__(self):
        return '{}(n_graphs={})'.format(self.__class__.__name__, self.n_graphs)

    @property
    def path(self):
        return osp.join(DATASET_FOLDER, self.__class__.__name__)

    @property
    def n_graphs(self):
        return self.__len__()

    @property
    def n_nodes(self):
        if len(self.graphs) == 1 or len(set([g.n_nodes for g in self.graphs])) == 1:
            return self.graphs[0].n_nodes
        else:
            return None

    @property
    def n_node_features(self):
        if len(self.graphs) == 1 or len(set([g.n_node_features for g in self.graphs])) == 1:
            return self.graphs[0].n_node_features
        else:
            return None

    @property
    def n_edge_features(self):
        if len(self.graphs) == 1 or len(set([g.n_edge_features for g in self.graphs])) == 1:
            return self.graphs[0].n_edge_features
        else:
            return None

    @property
    def n_labels(self):
        if len(self.graphs) == 1 or len(set([g.n_labels for g in self.graphs])) == 1:
            return self.graphs[0].n_labels
        else:
            return None

    @property
    def signature(self):
        """
        This property computes the signature of the dataset, which can be
        passed to `spektral.data.utils.to_tf_signature(signature)` to compute
        the TensorFlow signature. You can safely ignore this property unless
        you are creating a custom `Loader`.

        A signature consist of the TensorFlow TypeSpec, shape, and dtype of
        all characteristic matrices of the graphs in the Dataset. This is
        returned as a dictionary of dictionaries, with keys `x`, `a`, `e`, and
        `y` for the four main data matrices.

        Each sub-dictionary will have keys `spec`, `shape` and `dtype`.
        """
        signature = {}
        graph = self.graphs[0]  # This is always non-empty
        if graph.x is not None:
            signature['x'] = dict()
            signature['x']['spec'] = get_spec(graph.x)
            signature['x']['shape'] = (None, self.n_node_features)
            signature['x']['dtype'] = tf.as_dtype(graph.x.dtype)
        if graph.a is not None:
            signature['a'] = dict()
            signature['a']['spec'] = get_spec(graph.a)
            signature['a']['shape'] = (None, None)
            signature['a']['dtype'] = tf.as_dtype(graph.a.dtype)
        if graph.e is not None:
            signature['e'] = dict()
            signature['e']['spec'] = get_spec(graph.e)
            signature['e']['shape'] = (None, self.n_edge_features)
            signature['e']['dtype'] = tf.as_dtype(graph.e.dtype)
        if graph.y is not None:
            signature['y'] = dict()
            signature['y']['spec'] = get_spec(graph.y)
            signature['y']['shape'] = (self.n_labels,)
            signature['y']['dtype'] = tf.as_dtype(np.array(graph.y).dtype)
        return signature



from tensorflow.keras.layers import Layer

from spektral.utils.keras import is_layer_kwarg, deserialize_kwarg, is_keras_kwarg, serialize_kwarg


class Pool(Layer):
    r"""
    A general class for pooling layers.

    You can extend this class to create custom implementations of pooling layers.

    Any extension of this class must implement the `call(self, inputs)` and
    `config(self)` methods.

    **Arguments**:

    - ``**kwargs`: additional keyword arguments specific to Keras' Layers, like
    regularizers, initializers, constraints, etc.
    """
    def __init__(self, **kwargs):
        super().__init__(**{k: v for k, v in kwargs.items() if is_keras_kwarg(k)})
        self.kwargs_keys = []
        for key in kwargs:
            if is_layer_kwarg(key):
                attr = kwargs[key]
                attr = deserialize_kwarg(key, attr)
                self.kwargs_keys.append(key)
                setattr(self, key, attr)

    def build(self, input_shape):
        self.built = True

    def call(self, inputs):
        raise NotImplementedError

    def get_config(self):
        base_config = super().get_config()
        keras_config = {}
        for key in self.kwargs_keys:
            keras_config[key] = serialize_kwarg(key, getattr(self, key))
        return {**base_config, **keras_config, **self.config}

    @property
    def config(self):
        return {}



from spektral.utils import gcn_filter


class GCNFilter(object):
    r"""
    Normalizes the adjacency matrix as described by
    [Kipf & Welling (2017)](https://arxiv.org/abs/1609.02907):
$$
    \A \leftarrow \hat\D^{-\frac{1}{2}} (\A + \I) \hat\D^{-\frac{1}{2}}
    $$

    where \( \hat\D_{ii} = 1 + \sum\limits_{j = 1}^{N} \A_{ij} \).

    **Arguments**

    - `symmetric`: If False, then it computes \(\hat\D^{-1} (\A + \I)\) instead.
    """
    def __init__(self, symmetric=True):
        self.symmetric = symmetric

    def __call__(self, graph):
        if graph.a is not None:
            graph.a = gcn_filter(graph.a, self.symmetric)

        return graph



import tensorflow as tf
from tensorflow.keras import backend as K

from spektral.layers import ops
from spektral.layers.convolutional.message_passing import MessagePassing


class AGNNConv(MessagePassing):
    r"""
    An Attention-based Graph Neural Network (AGNN) from the paper

    > [Attention-based Graph Neural Network for Semi-supervised Learning](https://arxiv.org/abs/1803.03735)<br>
    > Kiran K. Thekumparampil et al.

    **Mode**: single, disjoint.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \X' = \P\X
    $$
    where
    $$
        \P_{ij} = \frac{
            \exp \left( \beta \cos \left( \x_i, \x_j \right) \right)
        }{
            \sum\limits_{k \in \mathcal{N}(i) \cup \{ i \}}
            \exp \left( \beta \cos \left( \x_i, \x_k \right) \right)
        }
    $$
    and \(\beta\) is a trainable parameter.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input.

    **Arguments**

    - `trainable`: boolean, if True, then beta is a trainable parameter.
    Otherwise, beta is fixed to 1;
    - `activation`: activation function;
    """

    def __init__(self, trainable=True, aggregate='sum', activation=None, **kwargs):
        super().__init__(aggregate=aggregate, activation=activation, **kwargs)
        self.trainable = trainable

    def build(self, input_shape):
        assert len(input_shape) >= 2
        if self.trainable:
            self.beta = self.add_weight(shape=(1,), initializer='ones', name='beta')
        else:
            self.beta = K.constant(1.)
        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        x_norm = K.l2_normalize(x, axis=-1)
        output = self.propagate(x, a, x_norm=x_norm)
        output = self.activation(output)

        return output

    def message(self, x, x_norm=None):
        x_j = self.get_j(x)
        x_norm_i = self.get_i(x_norm)
        x_norm_j = self.get_j(x_norm)
        alpha = self.beta * tf.reduce_sum(x_norm_i * x_norm_j, axis=-1)
        alpha = ops.unsorted_segment_softmax(alpha, self.index_i, self.n_nodes)
        alpha = alpha[:, None]

        return alpha * x_j

    @property
    def config(self):
        return {
            'trainable': self.trainable,
        }



import os
import os.path as osp

import networkx as nx
import numpy as np
import requests
import scipy.sparse as sp

from spektral.data import Dataset, Graph
from spektral.datasets.utils import DATASET_FOLDER
from spektral.utils.io import load_binary


class Citation(Dataset):
    """
    The citation datasets Cora, Citeseer and Pubmed.

    Node attributes are bag-of-words vectors representing the most common words
    in the text document associated to each node.
    Two papers are connected if either one cites the other.
    Labels represent the subject area of the paper.

    The train, test, and validation splits are given as binary masks and are
    accessible via the `mask_tr`, `mask_va`, and `mask_te` attributes.

    **Arguments**

    - `name`: name of the dataset to load (`'cora'`, `'citeseer'`, or
    `'pubmed'`);
    - `random_split`: if True, return a randomized split (20 nodes per class
    for training, 30 nodes per class for validation and the remaining nodes for
    testing, as recommended by [Shchur et al. (2018)](https://arxiv.org/abs/1811.05868)).
    If False (default), return the "Planetoid" public splits defined by
    [Yang et al. (2016)](https://arxiv.org/abs/1603.08861).
    - `normalize_x`: if True, normalize the features.
    """
    url = 'https://github.com/tkipf/gcn/raw/master/gcn/data/{}'
    suffixes = ['x', 'y', 'tx', 'ty', 'allx', 'ally', 'graph', 'test.index']

    def __init__(self, name, random_split=False, normalize_x=False, **kwargs):
        self.name = name.lower()
        if self.name not in self.available_datasets:
            raise ValueError('Unknown dataset {}. See Citation.available_datasets '
                             'for a list of available datasets.')
        self.random_split = random_split
        self.normalize_x = normalize_x
        self.mask_tr = self.mask_va = self.mask_te = None
        super().__init__(**kwargs)

    @property
    def path(self):
        return osp.join(DATASET_FOLDER, 'Citation', self.name)

    def read(self):
        objects = [_read_file(self.path, self.name, s) for s in self.suffixes]
        objects = [o.A if sp.issparse(o) else o for o in objects]
        x, y, tx, ty, allx, ally, graph, idx_te = objects

        # Public Planetoid splits. This is the default
        idx_tr = np.arange(y.shape[0])
        idx_va = np.arange(y.shape[0], y.shape[0] + 500)
        idx_te = idx_te.astype(int)
        idx_te_sort = np.sort(idx_te)

        # Fix disconnected nodes in Citeseer
        if self.name == 'citeseer':
            idx_te_len = idx_te.max() - idx_te.min() + 1
            tx_ext = np.zeros((idx_te_len, x.shape[1]))
            tx_ext[idx_te_sort - idx_te.min(), :] = tx
            tx = tx_ext
            ty_ext = np.zeros((idx_te_len, y.shape[1]))
            ty_ext[idx_te_sort - idx_te.min(), :] = ty
            ty = ty_ext

        x = np.vstack((allx, tx))
        y = np.vstack((ally, ty))
        x[idx_te, :] = x[idx_te_sort, :]
        y[idx_te, :] = y[idx_te_sort, :]

        # Row-normalize the features
        if self.normalize_x:
            print('Pre-processing node features')
            x = _preprocess_features(x)

        if self.random_split:
            # Throw away public splits and compute random ones like Shchur et al.
            from sklearn.model_selection import train_test_split
            indices = np.arange(y.shape[0])
            n_classes = y.shape[1]
            idx_tr, idx_te, y_tr, y_te = train_test_split(
                indices, y, train_size=20 * n_classes, stratify=y)
            idx_va, idx_te, y_va, y_te = train_test_split(
                idx_te, y_te, train_size=30 * n_classes, stratify=y_te)

        # Adjacency matrix
        a = nx.adjacency_matrix(nx.from_dict_of_lists(graph))  # CSR
        a.setdiag(0)
        a.eliminate_zeros()

        # Train/valid/test masks
        self.mask_tr = _idx_to_mask(idx_tr, y.shape[0])
        self.mask_va = _idx_to_mask(idx_va, y.shape[0])
        self.mask_te = _idx_to_mask(idx_te, y.shape[0])

        return [Graph(x=x, a=a, y=y)]

    def download(self):
        print('Downloading {} dataset.'.format(self.name))
        os.makedirs(self.path, exist_ok=True)
        for n in self.suffixes:
            f_name = 'ind.{}.{}'.format(self.name, n)
            req = requests.get(self.url.format(f_name))
            if req.status_code == 404:
                raise ValueError('Cannot download dataset ({} returned 404).'
                                 .format(self.url.format(f_name)))
            with open(os.path.join(self.path, f_name), 'wb') as out_file:
                out_file.write(req.content)

    @property
    def available_datasets(self):
        return ['cora', 'citeseer', 'pubmed']


class Cora(Citation):
    """
    Alias for `Citation('cora')`.
    """
    def __init__(self, random_split=False, normalize_x=False, **kwargs):
        super().__init__('cora', random_split=random_split,
                         normalize_x=normalize_x, **kwargs)


class Citeseer(Citation):
    """
    Alias for `Citation('citeseer')`.
    """
    def __init__(self, random_split=False, normalize_x=False, **kwargs):
        super().__init__('citeseer', random_split=random_split,
                         normalize_x=normalize_x, **kwargs)


class Pubmed(Citation):
    """
    Alias for `Citation('pubmed')`.
    """
    def __init__(self, random_split=False, normalize_x=False, **kwargs):
        super().__init__('pubmed', random_split=random_split,
                         normalize_x=normalize_x, **kwargs)


def _read_file(path, name, suffix):
    full_fname = os.path.join(path, 'ind.{}.{}'.format(name, suffix))
    if suffix == 'test.index':
        return np.loadtxt(full_fname)
    else:
        return load_binary(full_fname)


def _idx_to_mask(idx, l):
    mask = np.zeros(l)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)


def _preprocess_features(features):
    rowsum = np.array(features.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    features = r_mat_inv.dot(features)
    return features
